/*
 * ===========================================================================================
 *
 *       filename:  interrupt.c
 *
 *    description:  interrupt handler procedure.
 *
 *        version:  melis3.0
 *         create:  2020-06-29 19:35:50
 *       revision:  none
 *       compiler:  gcc:version 7.2.1 20170904 (release),arm/embedded-7-branch revision 255204
 *
 *         author:  caozilong@allwinnertech.com
 *   organization:  bu1-psw
 *  last modified:  2020-06-30 11:10:34
 *
 * ===========================================================================================
 */
#include <stddef.h>
#include <stdint.h>
#include <excep.h>
#include <stdio.h>
#include <irqflags.h>
#include <debug.h>
#include <csr.h>
#include <log.h>

#ifdef CONFIG_DEBUG_BACKTRACE
#include <backtrace.h>
#endif

extern uint8_t melis_kernel_running;
void awos_arch_save_fpu_status(fpu_context_t *);
void awos_arch_restore_fpu_status(fpu_context_t *);

typedef unsigned long (* exception_callback)(
    char *str,
    unsigned long scause,
	unsigned long sepc,
	unsigned long stval,
	irq_regs_t *regs);

typedef struct _exception_handle_table_entry
{
    int cause;
    exception_callback callback;
    char *error_str;
} exception_handle_table_entry;

exception_handle_table_entry exception_table[] =
{
    {
        .cause = EXC_INST_MISALIGNED,
        .error_str = "EXC_INST_MISALIGNED",
    },
    {
        .cause = EXC_INST_ACCESS,
        .error_str = "EXC_INST_ACCESS",
    },
    {
        .cause = EXC_INST_ILLEGAL,
        .error_str = "EXC_INST_ILLEGAL",
    },
    {
        .cause = EXC_BREAKPOINT,
        .error_str = "EXC_BREAKPOINT",
    },
    {
        .cause = EXC_LOAD_MISALIGN,
        .error_str = "EXC_LOAD_MISALIGN",
    },
    {
        .cause = EXC_LOAD_ACCESS,
        .error_str = "EXC_LOAD_ACCESS",
    },
    {
        .cause = EXC_STORE_MISALIGN,
        .error_str = "EXC_STORE_MISALIGN",
    },
    {
        .cause = EXC_STORE_ACCESS,
        .error_str = "EXC_STORE_ACCESS",
    },
    {
        .cause = EXC_INST_PAGE_FAULT,
        .error_str = "EXC_INST_PAGE_FAULT",
    },
    {
        .cause = EXC_LOAD_PAGE_FAULT,
        .error_str = "EXC_LOAD_PAGE_FAULT",
    },
    {
        .cause = EXC_STORE_PAGE_FAULT,
        .error_str = "EXC_STORE_PAGE_FAULT",
    },
    {
        .cause = EXC_SYSCALL_FRM_U,
        .error_str = "EXC_SYSCALL_FRM_U",
    },
    {
        .cause = EXC_SYSCALL_FRM_M,
        .error_str = "EXC_SYSCALL_FRM_M",
    },
};

static int check_fpu_status_clean(void)
{
    unsigned long sstatus;

    sstatus = arch_local_save_flags();
    if ((sstatus & SR_FS) != SR_FS_CLEAN)
    {
        return 0;
    }
    return 1;
}

void fpu_save_inirq(unsigned long sstatus)
{
#if 0
    rt_thread_t self = rt_thread_self();

    if (!melis_kernel_running)
    {
        return;
    }

    if (self == RT_NULL)
    {
        software_break();
    }

    if ((sstatus & SR_FS) == SR_FS_DIRTY)
    {
        awos_arch_save_fpu_status(&self->fdext_ctx);
    }

    if (!check_fpu_status_clean())
    {
        software_break();
    }
#endif
}

void fpu_restore_inirq(unsigned long sstatus)
{
#if 0
    rt_thread_t self = rt_thread_self();

    if (!melis_kernel_running)
    {
        return;
    }

    if (self == RT_NULL)
    {
        software_break();
    }

    // interrupt handler vector operations is integrity.
    // and has finished, so need not save its context.
    if ((sstatus & SR_FS) == SR_FS_DIRTY)
    {
        awos_arch_restore_fpu_status(&self->fdext_ctx);
    }

    if (!check_fpu_status_clean())
    {
        software_break();
    }
#endif
}

static void show_register(irq_regs_t *regs)
{
}

static unsigned long default_handle(char *str, unsigned long scause, unsigned long sepc, unsigned long stval, irq_regs_t *regs)
{
    printf("=====================================================================================================\n");
    printf("                                         %s              \n", str);
    printf("=====================================================================================================\n");
    show_register(regs);

   printf("\r\n");

#ifdef CONFIG_DEBUG_BACKTRACE
    printf("-------backtrace-----------\n");
    backtrace_exception(printk, regs->sstatus, regs->x2, regs->sepc, regs->x1);
    printf("---------------------------\n");
#endif

    void dump_system_information(void);
    dump_system_information();

    void dump_memory(unsigned long *buf, int len);
    dump_memory((unsigned long *)regs->x2, 128);

    while(1);
    return 0;
}

static unsigned long error_handle(unsigned long scause, unsigned long sepc, unsigned long stval, irq_regs_t *regs)
{
    int i;
    int cause = scause & ~SCAUSE_IRQ_FLAG;

    for(i = 0; i < sizeof(exception_table) / sizeof(exception_table[0]); i++)
    {
        if (exception_table[i].cause == cause && exception_table[i].error_str)
        {
            if (exception_table[i].callback)
            {
                return exception_table[i].callback(exception_table[i].error_str, scause, sepc, stval, regs);
            }
            else
            {
                return default_handle(exception_table[i].error_str, scause, sepc, stval, regs);
            }
        }
    }
    return default_handle("UNKNOWN ERROR", scause, sepc, stval, regs);
}

extern unsigned long esSYSCALL_function(irq_regs_t *regs, void *reserved);
unsigned long riscv_cpu_handle_exception(unsigned long scause, unsigned long sepc, unsigned long stval, irq_regs_t *regs)
{
    unsigned long ret = 0;

    if (scause & SCAUSE_IRQ_FLAG)
    {
	return default_handle("UNKNOWN ERROR", scause, sepc, stval, regs);
    }

    switch (scause & ~SCAUSE_IRQ_FLAG)
    {
        case EXC_INST_MISALIGNED:
        case EXC_INST_ACCESS:
        case EXC_INST_ILLEGAL:
        case EXC_BREAKPOINT:
        case EXC_LOAD_MISALIGN:
        case EXC_LOAD_ACCESS:
        case EXC_STORE_MISALIGN:
        case EXC_STORE_ACCESS:
        case EXC_SYSCALL_FRM_U:
        case EXC_SYSCALL_FRM_M:
        case EXC_INST_PAGE_FAULT:
        case EXC_LOAD_PAGE_FAULT:
        case EXC_STORE_PAGE_FAULT:
            return error_handle(scause, sepc, stval, regs);
        case EXC_SYSCALL_FRM_S:
            __inf("ecall SYSCALL_FROM_S");
            return esSYSCALL_function(regs, NULL);
        default:
            return default_handle("UNKNOWN ERROR", scause, sepc, stval, regs);
    }

    return 0;
}

extern void handle_exception(void);
void trap_init(void)
{
    unsigned long mie;
    /*
     * Set  scratch register to 0, indicating to exception vector
     * that we are presently executing in the kernel
     */
    csr_write(CSR_MSCRATCH, 0);

    /* Set the exception vector address */
    csr_write(CSR_MTVEC, &handle_exception);

    // disable S/M IPI, S timer and S external Interrupt.
    mie = 0;
    mie &= (~SIE_SSIE);
    mie &= (~MIE_MSIE);
    mie &= (~SIE_STIE);
    mie &= (~SIE_SEIE);

    // Enable M-Mode Timer and External Interrupt.
    mie |= (MIE_MTIE);
    mie |= (MIE_MEIE);
    /* Enable all interrupts */
    csr_write(CSR_MIE, mie);
}
