/*
 * ===========================================================================================
 *
 *       Filename:  mmu-sv39.c
 *
 *    Description:  mmu procedure.
 *
 *        Version:  Melis3.0
 *         Create:  2020-07-06 16:13:17
 *       Revision:  none
 *       Compiler:  GCC:version 7.2.1 20170904 (release),ARM/embedded-7-branch revision 255204
 *
 *         Author:  caozilong@allwinnertech.com
 *   Organization:  BU1-PSW
 *  Last Modified:  2020-07-08 14:24:21
 *
 * ===========================================================================================
 */
#include "page.h"
#include "pgtable.h"
#include <typedef.h>
#include <debug.h>
#include <log.h>

pgd_t system_pg_dir[PTRS_PER_PGD] __aligned(PAGE_SIZE);

// physical memory start address is 1G, and with size 64M.
// VA=PA memory map.
static unsigned long physical_memory_start_address = 0x40000000UL;
static unsigned long virtual_memory_start_address  = 0x40000000UL;
static unsigned long physical_memory_size          = 0x04000000UL;

static void mmu_create_map_region(pgd_t *pgdir, unsigned long va, unsigned long pa, unsigned long size, pgprot_t prot)
{
    unsigned long pgd_index = pgd_index(va);

    if (size != PGDIR_SIZE)
    {
        __err("direct map area must be 1G size.");
        software_break();
    }

    if (pgd_val(pgdir[pgd_index]) == 0)
    {
        pgdir[pgd_index] = pfn_pgd(PFN_DOWN(pa), prot);
    }
    else
    {
        __err("already maped on pgdir index %ld.", pgd_index);
        software_break();
    }

    return;
}

static void clear_pgd(void)
{
    int i = 0;

    pgd_t nullpgd = {0};
    for (i = 0; i < PTRS_PER_PGD; i ++)
    {
        system_pg_dir[i] = nullpgd;
    }
}

/*
 *IO  Attr: Strong order, uncached, unbufferd, shared, non-security.
 *Mem Attr: Weak   order, cached,   bufferd,   shared, non-security.
 *----------------------------------------------------------------------------
 *|0x4027f000 <system_pg_dir>:     0x90000000000000e7      0x70000000100000ef|
 *|0x4027f010 <system_pg_dir+16>:  0x70000000200000ef      0x0000000000000000|
 *----------------------------------------------------------------------------
 */
void mmu_init(void)
{
    clear_pgd();

    //[0G 2G] are directly map, VA=PA.
    //[2G 4G] are dynamic  map, VA!=PA.

    // crerate[0G 1G] map, IO area strong order, unbufferd, uncached..
    mmu_create_map_region(system_pg_dir, 0UL, 0UL, \
                          PGDIR_SIZE, \
                          pgprot_noncached(PAGE_KERNEL));


    // crerate[1G 2G] map, common memory area, weak order, buffered, cached.
    mmu_create_map_region(system_pg_dir, virtual_memory_start_address, \
                          physical_memory_start_address, \
                          physical_memory_size < PGDIR_SIZE ? PGDIR_SIZE : physical_memory_size, \
                          PAGE_KERNEL_EXEC);

    // crerate[2G 3G] map, common memory area, weak order, buffered, cached.
    mmu_create_map_region(system_pg_dir, 0x80000000UL, \
                          0x80000000UL, \
                          PGDIR_SIZE, \
                          PAGE_KERNEL_EXEC);

    //[3G-4G] areas leave for hole, are used for vmalloc dynamic map.
    return;
}

extern unsigned long awos_arch_virt_to_phys(unsigned long virtaddr);

unsigned long awos_arch_vaddr_map2_phys(unsigned long vaddr)
{
    unsigned long phyadr;

    if (vaddr == 0)
    {
        return 0;
    }

    //linear map zone, just cal with the offset.
    if ((vaddr >= virtual_memory_start_address) && (vaddr < virtual_memory_start_address + physical_memory_size))
    {
        return awos_arch_virt_to_phys(vaddr);
    }
    //__log("[%s %d]:the vaddr[0x%X] is not within the slab arrange!\r\n", __FUNCTION__,__LINE__,vaddr);

    /*modules runing address is virtual address which mapping to slab memory through mmu*/
    phyadr = 0;
    {
        pgd_t *pgd = NULL;
        p4d_t *p4d = NULL;
        pud_t *pud = NULL;
        pmd_t *pmd = NULL;
        pte_t *pte = NULL;

        pgd = pgd_offset_k(vaddr);
        if(pgd_none(*pgd) || !pgd_present(*pgd))
        {
            __inf("the [%p]'s pgd table item [%p] is none or invalid!\r\n", vaddr, pgd_val(*pgd));
            goto __out;
        }
        else
        {
            /* the real 1G item is in pud level, mmuv39 is only 3 levels, the 5 level is code structure */
            if(pgd_val(*pgd) & (_PAGE_READ | _PAGE_EXEC))
            {
                phyadr  = __pa((unsigned long)pgd_page_vaddr(*pgd) + (vaddr & (PGDIR_SIZE - 1)));
                __inf("the [%p]'s pgd level physic address is 0x%p!\r\n", vaddr, phyadr);
                goto __out;
            }
        }

        p4d = p4d_offset(pgd, vaddr);
        if(p4d_none(*p4d) || !p4d_present(*p4d))
        {
            __inf("the [%p]'s p4d table item [%p] is none or invalid!\r\n", vaddr, p4d_val(*p4d));
            goto __out;
        }

        pud = pud_offset(p4d, vaddr);
        if(pud_none(*pud) || !pud_present(*pud))
        {
            __inf("the [%p]'s pud table item [%p] is none or invalid!\r\n", vaddr, pud_val(*pud));
            goto __out;
        }

        pmd = pmd_offset(pud, vaddr);
        if(pmd_none(*pmd) || !pmd_present(*pmd))
        {
            __inf("the [%p]'s pmd table item [%p] is none or invalid!\r\n", vaddr, pmd_val(*pmd));
            goto __out;
        }
        else
        {
            if(pmd_val(*pmd) & (_PAGE_READ | _PAGE_EXEC))
            {
                phyadr  = __pa((unsigned long)pmd_page_vaddr(*pmd) + (vaddr & (PMD_SIZE - 1)));
                __inf("the [%p]'s pmd level physic address is %p!\r\n", vaddr, phyadr);
                goto __out;
            }
        }

        pte = pte_offset_kernel(pmd, vaddr);
        if(pte_none(*pte) || !pte_present(*pte))
        {
            __inf("the [%p]'s pte table item [%p] is none or invalid!\r\n", vaddr, pte_val(*pte));
            goto __out;
        }
        else
        {
            if(pte_val(*pte) & (_PAGE_READ | _PAGE_EXEC))
            {
                phyadr  = __pa((unsigned long)pte_page(*pte) + (vaddr & (PAGE_SIZE - 1)));
                __inf("the [%p]'s pte level physic address is %p!\r\n", vaddr, phyadr);
                goto __out;
            }
        }
        __err("if this log present, maybe need goto checking the vaddr request procedure!\r\n");
    }

__out:
    return phyadr;
}

/*
*********************************************************************************************************
*                                       CHECK VIRTUAL ADDRESS
*
* Description: check virtual address.
*
* Arguments  : vaddr     : the address need to be check in mmu table
*
* Returns    : 1 --- valid ; 0 --- invalid ;
*
* Note       : level by level to check the item's present bit until the last pte
*********************************************************************************************************
*/
int32_t check_virtual_address(unsigned long vaddr)
{
    if(0 == awos_arch_vaddr_map2_phys(vaddr))
    {
        return 0;
    }
    return 1;
}
